/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
 *
 * Copyright (C) 2008 David Zeuthen <david@fubar.dk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#define _GNU_SOURCE 1

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/utsname.h>
#include <net/if_arp.h>
#include <fcntl.h>
#include <signal.h>


#include <glib.h>
#include <glib-object.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include "devkit-daemon.h"
#include "devkit-daemon-glue.h"
#include "devkit-marshal.h"

/*--------------------------------------------------------------------------------------------------------------*/

enum
{
        DEVICE_EVENT_SIGNAL,
        LAST_SIGNAL,
};

static guint signals[LAST_SIGNAL] = { 0 };

struct DevkitDaemonPrivate
{
        DBusGConnection   *system_bus_connection;
        DBusGProxy        *system_bus_proxy;

	int                udev_socket;
	GIOChannel        *udev_channel;

        GList             *inhibitors;
        guint              killtimer_id;
        int                num_local_inhibitors;
        gboolean           no_exit;
};

static void     devkit_daemon_class_init  (DevkitDaemonClass *klass);
static void     devkit_daemon_init        (DevkitDaemon      *seat);
static void     devkit_daemon_finalize    (GObject     *object);

G_DEFINE_TYPE (DevkitDaemon, devkit_daemon, G_TYPE_OBJECT)

#define DEVKIT_DAEMON_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), DEVKIT_TYPE_DAEMON, DevkitDaemonPrivate))

typedef struct
{
        char *cookie;
        char *system_bus_name;
} Inhibitor;

static void
inhibitor_free (Inhibitor *inhibitor)
{
        g_free (inhibitor->cookie);
        g_free (inhibitor->system_bus_name);
        g_free (inhibitor);
}

static void
inhibitor_list_changed (DevkitDaemon *daemon)
{
        devkit_daemon_reset_killtimer (daemon);
}

void
devkit_daemon_inhibit_killtimer (DevkitDaemon *daemon)
{
        daemon->priv->num_local_inhibitors++;
        devkit_daemon_reset_killtimer (daemon);
}

void
devkit_daemon_uninhibit_killtimer (DevkitDaemon *daemon)
{
        daemon->priv->num_local_inhibitors--;
        devkit_daemon_reset_killtimer (daemon);
}

void main_quit (void);

static gboolean
killtimer_do_exit (gpointer user_data)
{
        g_debug ("Exiting due to inactivity");
        main_quit ();
        return FALSE;
}

void
devkit_daemon_reset_killtimer (DevkitDaemon *daemon)
{
        /* remove existing timer */
        if (daemon->priv->killtimer_id > 0) {
                g_debug ("Removed existing killtimer");
                g_source_remove (daemon->priv->killtimer_id);
                daemon->priv->killtimer_id = 0;
        }

        /* someone on the bus is inhibiting us */
        if (g_list_length (daemon->priv->inhibitors) > 0)
                return;

        /* some local long running method is inhibiting us */
        if (daemon->priv->num_local_inhibitors > 0)
                return;

        if (daemon->priv->no_exit)
                return;

        g_debug ("Setting killtimer to 30 seconds");
        daemon->priv->killtimer_id = g_timeout_add (30 * 1000, killtimer_do_exit, NULL);
}

static DBusHandlerResult
_filter (DBusConnection *connection, DBusMessage *message, void *user_data)
{
        DevkitDaemon *daemon = DEVKIT_DAEMON (user_data);
        const char *interface;

        interface = dbus_message_get_interface (message);

        if (dbus_message_is_signal (message, DBUS_INTERFACE_DBUS, "NameOwnerChanged")) {
                GList *l;
                const char *system_bus_name;
                const char *old_service_name;
                const char *new_service_name;

                /* find the name */
		if (dbus_message_get_args (message, NULL,
                                           DBUS_TYPE_STRING, &system_bus_name,
                                           DBUS_TYPE_STRING, &old_service_name,
                                           DBUS_TYPE_STRING, &new_service_name,
                                           DBUS_TYPE_INVALID)) {

                        if (strlen (new_service_name) == 0) {
                                /* he exited.. remove from inhibitor list */
                                for (l = daemon->priv->inhibitors; l != NULL; l = l->next) {
                                        Inhibitor *inhibitor = l->data;
                                        if (strcmp (system_bus_name, inhibitor->system_bus_name) == 0) {
                                                daemon->priv->inhibitors = g_list_remove (daemon->priv->inhibitors,
                                                                                          inhibitor);
                                                inhibitor_list_changed (daemon);
                                                g_debug ("removed inhibitor %s %s (disconnected from the bus)",
                                                         inhibitor->cookie, inhibitor->system_bus_name);
                                                break;
                                        }
                                }
                        }
                }
        }

        /* other filters might want to process this message too */
        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

/*--------------------------------------------------------------------------------------------------------------*/

GQuark
devkit_daemon_error_quark (void)
{
        static GQuark ret = 0;

        if (ret == 0) {
                ret = g_quark_from_static_string ("devkit_daemon_error");
        }

        return ret;
}


#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType
devkit_daemon_error_get_type (void)
{
        static GType etype = 0;

        if (etype == 0)
        {
                static const GEnumValue values[] =
                        {
                                ENUM_ENTRY (DEVKIT_DAEMON_ERROR_GENERAL, "GeneralError"),
                                ENUM_ENTRY (DEVKIT_DAEMON_ERROR_NOT_SUPPORTED, "NotSupported"),
                                ENUM_ENTRY (DEVKIT_DAEMON_ERROR_NO_SUCH_DEVICE, "NoSuchDevice"),
                                { 0, 0, 0 }
                        };
                g_assert (DEVKIT_DAEMON_NUM_ERRORS == G_N_ELEMENTS (values) - 1);
                etype = g_enum_register_static ("DevkitDaemonError", values);
        }
        return etype;
}

static void
devkit_daemon_class_init (DevkitDaemonClass *klass)
{
        GObjectClass   *object_class = G_OBJECT_CLASS (klass);

        object_class->finalize = devkit_daemon_finalize;

        g_type_class_add_private (klass, sizeof (DevkitDaemonPrivate));

        signals[DEVICE_EVENT_SIGNAL] =
                g_signal_new ("device-event",
                              G_OBJECT_CLASS_TYPE (klass),
                              G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                              0,
                              NULL, NULL,
                              devkit_marshal_VOID__STRING_STRING_STRING_STRING_BOXED_BOXED,
                              G_TYPE_NONE,
                              6,
                              /* action */
                              G_TYPE_STRING,
                              /* subsystem */
                              G_TYPE_STRING,
                              /* native path */
                              G_TYPE_STRING,
                              /* device file */
                              G_TYPE_STRING,
                              /* device file symlinks */
                              dbus_g_type_get_collection ("GPtrArray", G_TYPE_STRING),
                              /* properties */
                              dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_STRING)
                        );


        dbus_g_object_type_install_info (DEVKIT_TYPE_DAEMON, &dbus_glib_devkit_daemon_object_info);

        dbus_g_error_domain_register (DEVKIT_DAEMON_ERROR,
                                      NULL,
                                      DEVKIT_DAEMON_TYPE_ERROR);
}

static void
devkit_daemon_init (DevkitDaemon *daemon)
{
        daemon->priv = DEVKIT_DAEMON_GET_PRIVATE (daemon);
        daemon->priv->udev_socket = -1;
}

static void
devkit_daemon_finalize (GObject *object)
{
        DevkitDaemon *daemon;

        g_return_if_fail (object != NULL);
        g_return_if_fail (DEVKIT_IS_DAEMON (object));

        daemon = DEVKIT_DAEMON (object);

        g_return_if_fail (daemon->priv != NULL);

        if (daemon->priv->system_bus_proxy != NULL)
                g_object_unref (daemon->priv->system_bus_proxy);

        if (daemon->priv->system_bus_connection != NULL)
                dbus_g_connection_unref (daemon->priv->system_bus_connection);

        if (daemon->priv->inhibitors != NULL) {
                g_list_foreach (daemon->priv->inhibitors, (GFunc) inhibitor_free, NULL);
                g_list_free (daemon->priv->inhibitors);
        }

        if (daemon->priv->killtimer_id > 0) {
                g_source_remove (daemon->priv->killtimer_id);
        }

        if (daemon->priv->udev_socket != -1)
                close (daemon->priv->udev_socket);

        if (unlink (PACKAGE_LOCALSTATE_DIR "/run/devkit/udev_socket") != 0) {
                g_warning ("Cannot delete udev socket file: %m");
        }

        if (daemon->priv->udev_channel != NULL)
                g_io_channel_unref (daemon->priv->udev_channel);

        G_OBJECT_CLASS (devkit_daemon_parent_class)->finalize (object);
}


static gboolean
receive_udev_data (GIOChannel *source, GIOCondition condition, gpointer user_data)
{
        DevkitDaemon *daemon = user_data;
	int fd;
	int retval;
	struct msghdr smsg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	struct ucred *cred;
	char cred_msg[CMSG_SPACE(sizeof(struct ucred))];
	char buf[4096];
	size_t bufpos = 0;
	const char *action;
	const char *devpath;
	const char *subsystem;

	memset(buf, 0x00, sizeof (buf));
	iov.iov_base = &buf;
	iov.iov_len = sizeof (buf);
	memset (&smsg, 0x00, sizeof (struct msghdr));
	smsg.msg_iov = &iov;
	smsg.msg_iovlen = 1;
	smsg.msg_control = cred_msg;
	smsg.msg_controllen = sizeof (cred_msg);

	fd = g_io_channel_unix_get_fd (source);

	retval = recvmsg (fd, &smsg, 0);
	if (retval <  0) {
		if (errno != EINTR)
			g_warning ("Unable to receive message: %m");
		goto out;
	}
	cmsg = CMSG_FIRSTHDR (&smsg);
	cred = (struct ucred *) CMSG_DATA (cmsg);

	if (cmsg == NULL || cmsg->cmsg_type != SCM_CREDENTIALS) {
		g_warning ("No sender credentials received, message ignored");
		goto out;
	}

	if (cred->uid != 0) {
		g_warning ("Sender uid=%d, message ignored", cred->uid);
		goto out;
	}

	if (!strstr(buf, "@/")) {
		g_warning ("invalid message format");
		goto out;
	}

        const char *device_file;
        GPtrArray *symlinks;
        GHashTable *properties;

        symlinks = g_ptr_array_new ();
        properties = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

        action = NULL;
        devpath = NULL;
        subsystem = NULL;
        device_file = NULL;

	while (bufpos < sizeof (buf)) {
		size_t keylen;
		char *key;

		key = &buf[bufpos];
		keylen = strlen(key);
		if (keylen == 0)
			break;
		bufpos += keylen + 1;

		if (strncmp (key, "ACTION=", 7) == 0) {
			action = key + 7;
		} else if (strncmp (key, "DEVPATH=", 8) == 0) {
                        devpath = key + 8;
		} else if (strncmp(key, "SUBSYSTEM=", 10) == 0) {
                        subsystem = key + 10;
		} else if (strncmp(key, "DEVNAME=", 8) == 0) {
                        device_file = key + 8;
		} else if (strncmp(key, "DEVLINKS=", 9) == 0) {
                        char **tokens;
                        int n;
                        tokens = g_strsplit (key + 9, " ", 0);
                        for (n = 0; tokens[n] != NULL; n++) {
                                g_ptr_array_add (symlinks, g_strdup (tokens[n]));
                        }
                        g_strfreev (tokens);
                } else {
                        char *value;
                        value = strchr (key, '=');
                        if (value == NULL) {
                                g_warning ("line '%s' is malformed; ignoring", key);
                        } else {
                                g_hash_table_insert (properties,
                                                     g_strndup (key, value - key),
                                                     g_strdup (value + 1));
                        }
                }
	}

	if (action != NULL && devpath != NULL && subsystem != NULL) {
                char *native_path;

                native_path = g_build_filename ("/sys", devpath, NULL);

                g_signal_emit (daemon, signals[DEVICE_EVENT_SIGNAL], 0,
                               action,
                               subsystem,
                               native_path,
                               device_file != NULL ? device_file : "",
                               symlinks,
                               properties,
                               NULL);

                g_free (native_path);
	} else {
                g_warning ("malformed message");
        }

        g_hash_table_destroy (properties);
        g_ptr_array_foreach (symlinks, (GFunc) g_free, NULL);
out:
	return TRUE;
}

static gboolean
register_daemon (DevkitDaemon *daemon)
{
	struct sockaddr_un saddr;
	socklen_t addrlen;
	const int on = 1;
        DBusConnection *connection;
        GError *error = NULL;
        DBusError dbus_error;

        error = NULL;
        daemon->priv->system_bus_connection = dbus_g_bus_get (DBUS_BUS_SYSTEM, &error);
        if (daemon->priv->system_bus_connection == NULL) {
                if (error != NULL) {
                        g_critical ("error getting system bus: %s", error->message);
                        g_error_free (error);
                }
                goto error;
        }
        connection = dbus_g_connection_get_connection (daemon->priv->system_bus_connection);

        dbus_g_connection_register_g_object (daemon->priv->system_bus_connection, "/",
                                             G_OBJECT (daemon));

        daemon->priv->system_bus_proxy = dbus_g_proxy_new_for_name (daemon->priv->system_bus_connection,
                                                                    DBUS_SERVICE_DBUS,
                                                                    DBUS_PATH_DBUS,
                                                                    DBUS_INTERFACE_DBUS);

        /* need to listen to NameOwnerChanged */
        dbus_error_init (&dbus_error);
	dbus_bus_add_match (dbus_g_connection_get_connection (daemon->priv->system_bus_connection),
			    "type='signal'"
			    ",interface='"DBUS_INTERFACE_DBUS"'"
			    ",sender='"DBUS_SERVICE_DBUS"'"
			    ",member='NameOwnerChanged'",
			    &dbus_error);
        if (dbus_error_is_set (&dbus_error)) {
                g_warning ("Cannot add match rule: %s: %s", dbus_error.name, dbus_error.message);
                dbus_error_free (&dbus_error);
                goto error;
        }
        if (!dbus_connection_add_filter (dbus_g_connection_get_connection (daemon->priv->system_bus_connection),
                                         _filter,
                                         daemon,
                                         NULL)) {
                g_warning ("Cannot add D-Bus filter: %s: %s", dbus_error.name, dbus_error.message);
                dbus_error_free (&dbus_error);
                goto error;
        }

        if (g_file_test (PACKAGE_LOCALSTATE_DIR "/run/devkit/udev_socket", G_FILE_TEST_EXISTS)) {
                if (unlink (PACKAGE_LOCALSTATE_DIR "/run/devkit/udev_socket") != 0) {
                        g_warning ("Cannot delete stale udev socket: %m");
                        goto error;
                }
        }

	/* setup socket for listening from messages from udev */
	memset (&saddr, 0x00, sizeof(saddr));
	saddr.sun_family = AF_LOCAL;
	/* socket path */
	strcpy (saddr.sun_path, PACKAGE_LOCALSTATE_DIR "/run/devkit/udev_socket");
	addrlen = offsetof (struct sockaddr_un, sun_path) + strlen(saddr.sun_path+1) + 1;
	daemon->priv->udev_socket = socket (AF_LOCAL, SOCK_DGRAM, 0);
	if (daemon->priv->udev_socket == -1) {
		g_warning ("Couldn't open udev event socket: %m");
                goto error;
	}

	if (bind (daemon->priv->udev_socket, (struct sockaddr *) &saddr, addrlen) < 0) {
		g_warning ("Error binding to udev event socket: %m");
                goto error;
	}
	/* enable receiving of the sender credentials */
	setsockopt (daemon->priv->udev_socket, SOL_SOCKET, SO_PASSCRED, &on, sizeof(on));
	daemon->priv->udev_channel = g_io_channel_unix_new (daemon->priv->udev_socket);
	g_io_add_watch (daemon->priv->udev_channel, G_IO_IN, receive_udev_data, daemon);
	g_io_channel_unref (daemon->priv->udev_channel);

        devkit_daemon_reset_killtimer (daemon);

        return TRUE;

error:
        return FALSE;
}


DevkitDaemon *
devkit_daemon_new (gboolean no_exit)
{
        gboolean res;
        DevkitDaemon *daemon;

        daemon = DEVKIT_DAEMON (g_object_new (DEVKIT_TYPE_DAEMON, NULL));
        daemon->priv->no_exit = no_exit;

        res = register_daemon (DEVKIT_DAEMON (daemon));
        if (! res) {
                g_object_unref (daemon);
                return NULL;
        }

        return daemon;
}

/*--------------------------------------------------------------------------------------------------------------*/

#if 0
static gboolean
throw_error (DBusGMethodInvocation *context, int error_code, const char *format, ...)
{
        GError *error;
        va_list args;
        char *message;

        va_start (args, format);
        message = g_strdup_vprintf (format, args);
        va_end (args);

        error = g_error_new (DEVKIT_DAEMON_ERROR,
                             error_code,
                             message);
        dbus_g_method_return_error (context, error);
        g_error_free (error);
        g_free (message);
        return TRUE;
}
#endif

/*--------------------------------------------------------------------------------------------------------------*/
/* exported methods */

gboolean
devkit_daemon_inhibit_shutdown (DevkitDaemon     *daemon,
                                      DBusGMethodInvocation *context)
{
        GList *l;
        Inhibitor *inhibitor;
        const char *system_bus_name;

        system_bus_name = dbus_g_method_get_sender (context);

        inhibitor = g_new0 (Inhibitor, 1);
        inhibitor->system_bus_name = g_strdup (system_bus_name);
regen_cookie:
        inhibitor->cookie = g_strdup_printf ("%d", g_random_int_range (0, G_MAXINT));
        for (l = daemon->priv->inhibitors; l != NULL; l = l->next) {
                Inhibitor *i = l->data;
                if (strcmp (i->cookie, inhibitor->cookie) == 0) {
                        g_free (inhibitor->cookie);
                        goto regen_cookie;
                }
        }

        daemon->priv->inhibitors = g_list_prepend (daemon->priv->inhibitors, inhibitor);
        inhibitor_list_changed (daemon);

        g_debug ("added inhibitor %s %s", inhibitor->cookie, inhibitor->system_bus_name);

        dbus_g_method_return (context, inhibitor->cookie);
        return TRUE;
}

/*--------------------------------------------------------------------------------------------------------------*/

gboolean
devkit_daemon_uninhibit_shutdown (DevkitDaemon     *daemon,
                                        const char            *cookie,
                                        DBusGMethodInvocation *context)
{
        GList *l;
        GError *error;
        const char *system_bus_name;

        system_bus_name = dbus_g_method_get_sender (context);

        for (l = daemon->priv->inhibitors; l != NULL; l = l->next) {
                Inhibitor *inhibitor = l->data;
                if (strcmp (cookie, inhibitor->cookie) == 0 &&
                    strcmp (system_bus_name, inhibitor->system_bus_name) == 0) {

                        daemon->priv->inhibitors = g_list_remove (daemon->priv->inhibitors, inhibitor);
                        inhibitor_list_changed (daemon);

                        g_debug ("removed inhibitor %s %s", inhibitor->cookie, inhibitor->system_bus_name);
                        dbus_g_method_return (context);
                        goto out;
                }
        }

        error = g_error_new (DEVKIT_DAEMON_ERROR,
                             DEVKIT_DAEMON_ERROR_GENERAL,
                             "No such inhibitor");
        dbus_g_method_return_error (context, error);
        g_error_free (error);

out:
        return TRUE;
}

/*--------------------------------------------------------------------------------------------------------------*/

#define STRUCT_TYPE (dbus_g_type_get_struct ("GValueArray",             \
                                             G_TYPE_STRING,             \
                                             G_TYPE_STRING,             \
                                             G_TYPE_STRING,             \
                                             G_TYPE_STRV,                                        \
                                             dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_STRING), \
                                             G_TYPE_INVALID))

static void
enumerate_add_device (const char *subsystem,
                      const char *native_path,
                      const char *device_file,
                      GPtrArray *device_file_symlinks,
                      GHashTable *properties,
                      GPtrArray *devices)
{
        GValue elem = {0};

        if (subsystem != NULL && native_path != NULL) {
                g_value_init (&elem, STRUCT_TYPE);
                g_value_take_boxed (&elem, dbus_g_type_specialized_construct (STRUCT_TYPE));
                dbus_g_type_struct_set (&elem,
                                        0, subsystem,
                                        1, native_path,
                                        2, device_file != NULL ? device_file : "",
                                        3, device_file_symlinks->pdata,
                                        4, properties,
                                        G_MAXUINT);
                g_ptr_array_add (devices, g_value_get_boxed (&elem));
        } else {
                g_warning ("either subsytem or native_path is NULL");
        }
}

static char *
_dupv8 (const char *s)
{
        const char *end_valid;

        if (!g_utf8_validate (s,
                             -1,
                             &end_valid)) {
                g_warning ("The string '%s' is not valid UTF-8. Invalid characters begins at '%s'", s, end_valid);
                return g_strndup (s, end_valid - s);
        } else {
                return g_strdup (s);
        }
}

static void
retrieve_data_from_udev (const char  *sysfs_path,
                         char       **device_file,
                         GPtrArray   *device_file_symlinks,
                         GHashTable  *properties)
{
        int n;
        char *s;
        const char *p;
        char *db_path;
        char *db_file_path;
        char *contents;
        char **lines;

        g_return_if_fail (g_str_has_prefix (sysfs_path, "/sys"));
        g_return_if_fail (device_file != NULL);
        g_return_if_fail (device_file_symlinks != NULL);
        g_return_if_fail (properties != NULL);

        db_file_path = NULL;
        db_path = g_new0 (char, 3 * (strlen (sysfs_path) - 4) + 1);

        for (p = sysfs_path + 4, n = 0; *p != '\0'; p++) {
                switch (*p) {
                case '/':
                        db_path[n++] = '\\';
                        db_path[n++] = 'x';
                        db_path[n++] = '2';
                        db_path[n++] = 'f';
                        break;
                case '\\':
                        db_path[n++] = '\\';
                        db_path[n++] = 'x';
                        db_path[n++] = '5';
                        db_path[n++] = 'c';
                        break;
                default:
                        db_path[n++] = *p;
                        break;
                }
        }
        db_path[n++] = '\0';

        db_file_path = g_build_filename ("/dev", ".udev", "db", db_path, NULL);

        if (g_file_test (db_file_path, G_FILE_TEST_IS_SYMLINK)) {
                char target[512];
                ssize_t target_len;

                target_len = readlink (db_file_path, target, sizeof target - 1);
                if (target_len == -1) {
                        g_warning ("Cannot resolve link %s: %m", db_file_path);
                        goto out;
                }
                target[target_len] = '\0';

                if (*device_file != NULL)
                        g_free (*device_file);
                s = _dupv8 (target);
                *device_file = g_build_filename ("/dev", s, NULL);
                g_free (s);

        } else {
                if (!g_file_get_contents (db_file_path, &contents, NULL, NULL)) {
                        /* not necessary fatal; there may not be anything in the database for this path */
                        goto out;
                }
                lines = g_strsplit (contents, "\n", 0);
                for (n = 0; lines[n] != NULL; n++) {
                        const char *line = lines[n];

                        if (g_str_has_prefix (line, "N:")) {
                                if (*device_file != NULL)
                                        g_free (*device_file);
                                s = _dupv8 (line + 2);
                                *device_file = g_build_filename ("/dev", s, NULL);
                                g_free (s);
                        } else if (g_str_has_prefix (line, "S:")) {
                                s = _dupv8 (line + 2);
                                g_ptr_array_add (device_file_symlinks, g_build_filename ("/dev", s, NULL));
                                g_free (s);
                        } else if (g_str_has_prefix (line, "E:")) {
                                char *key;
                                key = strchr (line + 2, '=');
                                if (key != NULL) {
                                        *key = '\0';
                                        g_hash_table_insert (properties, _dupv8 (line + 2), _dupv8 (key + 1));
                                } else {
                                        g_warning ("malformed property line '%s'", line + 2);
                                }
                        }
                }

                g_strfreev (lines);
                g_free (contents);
        }

out:
        g_free (db_file_path);
        g_free (db_path);
}

gboolean
devkit_daemon_enumerate_by_subsystem  (DevkitDaemon           *daemon,
                                       const char            **subsystems,
                                       DBusGMethodInvocation  *context)
{
        int n, m;
        GPtrArray *devices;

        devices = dbus_g_type_specialized_construct (dbus_g_type_get_collection ("GPtrArray", STRUCT_TYPE));

        for (n = 0; subsystems[n] != NULL; n++) {
                char *dirname;
                GDir *dir;
                const char *name;

                /* collect devices from both /sys/class and /sys/buss */
                for (m = 0; m < 2; m++) {
                        switch (m) {
                        case 0:
                                dirname = g_build_filename ("/sys/class", subsystems[n], NULL);
                                break;
                        case 1:
                                dirname = g_build_filename ("/sys/bus", subsystems[n], "devices", NULL);
                                break;
                        default:
                                g_assert_not_reached ();
                                break;
                        }

                        dir = g_dir_open (dirname, 0, NULL);
                        while (dir != NULL && (name = g_dir_read_name (dir)) != NULL) {
                                char *s;
                                char sysfs_path[PATH_MAX];
                                char *device_file;
                                GPtrArray *device_file_symlinks;
                                GHashTable *properties;

                                s = g_build_filename (dirname, name, NULL);
                                if (realpath (s, sysfs_path) == NULL) {
                                        g_warning ("path '%s' is malformed", s);
                                        g_free (s);
                                        continue;
                                }
                                g_free (s);

                                device_file = NULL;
                                device_file_symlinks = g_ptr_array_new ();
                                properties = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

                                retrieve_data_from_udev (sysfs_path, &device_file, device_file_symlinks, properties);

                                g_ptr_array_add (device_file_symlinks, NULL);

                                enumerate_add_device (subsystems[n],
                                                      sysfs_path,
                                                      device_file,
                                                      device_file_symlinks,
                                                      properties,
                                                      devices);

                                g_free (device_file);
                                g_ptr_array_foreach (device_file_symlinks, (GFunc) g_free, NULL);
                                g_ptr_array_free (device_file_symlinks, TRUE);
                                g_hash_table_destroy (properties);
                        }
                        if (dir != NULL)
                                g_dir_close (dir);
                }
                g_free (dirname);
        }

        dbus_g_method_return (context, devices);
        return TRUE;
}

gboolean
devkit_daemon_enumerate_by_native_path  (DevkitDaemon           *daemon,
                                         const char            **native_paths,
                                         DBusGMethodInvocation  *context)
{
        int n;
        GPtrArray *devices;
        char *s;
        char sysfs_path[PATH_MAX];
        char subsystem_path[PATH_MAX];
        char *device_file;
        char *subsystem;
        GPtrArray *device_file_symlinks;
        GHashTable *properties;
        ssize_t len;

        devices = dbus_g_type_specialized_construct (dbus_g_type_get_collection ("GPtrArray", STRUCT_TYPE));

        for (n = 0; native_paths[n] != NULL; n++) {
                device_file = NULL;
                subsystem = NULL;
                device_file_symlinks = g_ptr_array_new ();
                properties = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);

                if (realpath (native_paths[n], sysfs_path) == NULL) {
                        g_warning ("path '%s' is malformed", native_paths[n]);
                        continue;
                }

                s = g_build_filename (sysfs_path, "subsystem", NULL);
                len = readlink (s, subsystem_path, sizeof subsystem_path - 1);
                if (len < 0) {
                        g_warning ("Cannot resolve symlink '%s'", s);
                        g_free (s);
                        continue;
                }
                subsystem_path[len] = '\0';
                subsystem = g_path_get_basename (subsystem_path);
                g_free (s);

                retrieve_data_from_udev (sysfs_path, &device_file, device_file_symlinks, properties);
                g_ptr_array_add (device_file_symlinks, NULL);

                enumerate_add_device (subsystem,
                                      sysfs_path,
                                      device_file,
                                      device_file_symlinks,
                                      properties,
                                      devices);

                g_free (device_file);
                g_free (subsystem);
                g_ptr_array_foreach (device_file_symlinks, (GFunc) g_free, NULL);
                g_ptr_array_free (device_file_symlinks, TRUE);
                g_hash_table_destroy (properties);
        }

        dbus_g_method_return (context, devices);
        return TRUE;
}

/*--------------------------------------------------------------------------------------------------------------*/
